/*
 This is part of the OTF library. Copyright by ZIH, TU Dresden 2005-2006.
 Authors: Andreas Knuepfer, Holger Brunst, Ronny Brendel, Thomas Kriebitzsch
*/

#ifndef STATE_H
#define STATE_H

#include "config.h"

#include <inttypes.h>

#include <deque>
#include <map>
#include <list>


#include <otf.h>


/* *** function stuff *** ********************************* */


struct FunctionCall {


public:

	uint64_t time;
	uint32_t token;


public:

	/** constructor */
	FunctionCall( uint64_t t, uint32_t f ) : time( t ), token( f ) {};

	/** copy constructor */
	FunctionCall( const FunctionCall& other ) : time(other.time), token(other.token) {};
};


struct FunctionStatistics {


public:

	int64_t occurrences;
	int64_t exclusiveTime;
	int64_t inclusiveTime;


public:

	/** constructor */
	FunctionStatistics( int64_t o= 0, int64_t ex= 0, int64_t in= 0 ) : 
		occurrences( o ), exclusiveTime( ex ), inclusiveTime( in ) {};

	/** copy constructor */
	FunctionStatistics( const FunctionStatistics& other ) : 
		occurrences( other.occurrences ), exclusiveTime( other.exclusiveTime ), 
		inclusiveTime( other.inclusiveTime ) {};
};


/* *** send stuff ************************************** */
struct Send {


public:
	
	uint64_t originaltime;
	uint32_t receiver;
	uint32_t procGroup;
	uint32_t tag;
	uint32_t source;
	
public:

	Send( uint64_t ot, uint32_t r, uint32_t pg, uint32_t t, uint32_t s ) : 
		originaltime( ot ), receiver( r ), procGroup( pg ), tag( t ),
		source( s ) {};
		
	Send( const Send& other ) : originaltime( other.originaltime ),
		receiver( other.receiver ), procGroup( other.procGroup ),
		tag( other.tag ), source( other.source ) {};
};


struct SendStatistics {

public:
	
	uint64_t number_sent;
	uint64_t number_recvd;
	uint64_t bytes_sent;
	uint64_t bytes_recvd;
	
public:

	SendStatistics( uint64_t ns= 0, uint64_t nr= 0, uint64_t bs= 0, uint64_t br= 0 ) :
		number_sent( ns ), number_recvd( nr ), bytes_sent( bs ),
		bytes_recvd( br ) {};
	
	SendStatistics( const SendStatistics& other ) :
		number_sent( other.number_sent ), number_recvd( other.number_recvd ),
		bytes_sent( other.bytes_sent ), bytes_recvd( other.bytes_recvd ) {};
};



/* *** ProcessState *** ********************************* */


/** class containing the state of a process trace at a given */
class ProcessState {


public:

	/* function stack */
	std::deque<FunctionCall> fstack;
	
	/* list of pending messages */
	std::deque<Send> sstack;

	/* statistic per function since the beginning of the trace */
	std::map<uint32_t,FunctionStatistics> fstatistics;

	/* statistics for messages */
	SendStatistics sstatistics;


public:

	/** constructor */
	ProcessState() {};


	void enterFunction( uint64_t time, uint32_t token );

	void leaveFunction( uint64_t time, uint32_t token );

	void sendMessage( uint64_t time, uint32_t receiver, uint32_t procGroup,
		uint32_t tag, uint32_t msglength, uint32_t source );
	
	void recvMessage( uint32_t msglength );
		
	void matchMessage( uint32_t receiver, uint32_t procGroup, uint32_t tag );
	
	void printStack( uint32_t processid ) const;
	void printStatistics( uint32_t processid, uint64_t time,
		std::map< uint32_t, uint32_t> *functiongroups ) const;
	void printSends( uint32_t processid ) const;

	void writeStack( OTF_Writer* writer, uint64_t time, uint32_t processid ) const;
	void writeStatistics( OTF_Writer* writer, uint64_t time,
		uint32_t processid,
		std::map< uint32_t, uint32_t> *functiongroups ) const;
	void writeSends( OTF_Writer* writer, uint64_t time, uint32_t processid ) const;
};


/* *** State *** **************************************** */


/** state of a whole trace */
class State {


	uint32_t x;

	std::map<uint32_t,ProcessState> processes;

	/* maps the function to its funtiongroupid */
	std::map< uint32_t, uint32_t> functiongroups;
	
	bool usefunctiongroups;
	bool doSnapshots;
	bool doStatistics;

public:

	/** constructor */
	State( bool _usefunctiongroups, bool _doSnapshots= true, bool _doStatistics= true ) : 
		usefunctiongroups( _usefunctiongroups ), doSnapshots( _doSnapshots ), 
		doStatistics( _doStatistics ) {};


	void defProcess( uint32_t processid );

	void defFunction( uint32_t function, uint32_t group );

	void enterFunction( uint64_t time, uint32_t processid, uint32_t token );

	void leaveFunction( uint64_t time, uint32_t processid, uint32_t token );
	
	void sendMessage( uint64_t time, uint32_t sender, uint32_t receiver,
		uint32_t procGroup, uint32_t tag, uint32_t length, uint32_t source );
	
	void recvMessage( uint32_t sender, uint32_t receiver, uint32_t procGroup,
		uint32_t tag, uint32_t msglength );

	void printStack() const;
	void printStatistics( uint64_t time );
	void printSends() const;
	
	void writeSnapshot( OTF_Writer* writer, uint64_t time ) const;
	void writeStatistics( OTF_Writer* writer, uint64_t time );
};


#endif /* STATE_H */

