/*
 This is part of the OTF library. Copyright by ZIH, TU Dresden 2005-2006.
 Authors: Andreas Knuepfer, Holger Brunst, Ronny Brendel, Thomas Kriebitzsch
*/

#include "config.h"

#include "Handler.h"

#include <unistd.h>
#include <set>
#include <iostream>
#include <assert.h>
using namespace std;


#include <otf.h>


#define HELPTEXT "" \
" otfdump - convert otf traces or parts of it into a human readable, long  \n" \
"               version                                                    \n" \
"                                                                          \n" \
" Options:                                                                 \n" \
"     -h, --help     show this help message                                \n" \
"     -V             show OTF version                                      \n" \
"     -o <file>      output file                                           \n" \
"                    if the ouput file is unspecified the stdout will be   \n" \
"                    used                                                  \n" \
"     --num <a> <b>  output only records no. [a,b]                         \n" \
"     --time <a> <b> output only records with time stamp in [a,b]          \n" \
"     --nodef        omit definition records                               \n" \
"                                                                          \n" \
"                                                                          \n" \
" ToDo:                                                                    \n" \
"          - option for process selection                                  \n" \
"          - option for event type selection                               \n" \
"                                                                          \n" \


int main ( int argc, const char** argv ) {


	char* filename = NULL;
	int buffersize= 4*1024;

	OTF_FileManager* manager;
	OTF_Reader* reader= NULL;
	OTF_MasterControl* mc= NULL;

	OTF_HandlerArray* handlers;
	
	char *namestub;

	unsigned int maxfilehandles = 100;

	uint64_t minNum= (uint64_t) 0;
	uint64_t maxNum= (uint64_t) -1;
	uint64_t minTime= (uint64_t) 0;
	uint64_t maxTime= (uint64_t) -1;
	uint32_t nodef= 0;


	/* argument handling */

	if ( 1 >= argc ) {

		printf( HELPTEXT );
		exit(0);
	}

	for ( int i = 1; i < argc; i++ ) {

		if ( ( 0 == strcmp( "-b", argv[i] ) ) && ( i+1 < argc ) ) {
		
			buffersize = atoi( argv[i+1] );
			++i;

		} else if ( 0 == strcmp( "--help", argv[i] ) ||	0 == strcmp( "-h", argv[i] ) ) {

			printf( HELPTEXT );
			exit(0);

		} else if ( 0 == strcmp( "-V", argv[i] ) ) {

			printf( "%u.%u.%u \"%s\"\n", OTF_VERSION_MAYOR, OTF_VERSION_MINOR,
				OTF_VERSION_SUB, OTF_VERSION_STRING);
			exit( 0 );

		} else if ( ( 0 == strcmp( "-f", argv[i] ) ) && ( i+1 < argc ) ) {

			maxfilehandles= atoi( argv[i+1] );
			++i;

		} else if ( ( 0 == strcmp( "--num", argv[i] ) ) && ( i+2 < argc ) ) {

			minNum= atoll( argv[i+1] );
			maxNum= atoll( argv[i+2] );

			i += 2;

		} else if ( ( 0 == strcmp( "--time", argv[i] ) ) && ( i+2 < argc ) ) {

			minTime= atoll( argv[i+1] );
			maxTime= atoll( argv[i+2] );

			i += 2;

		} else if ( 0 == strcmp( "--nodef", argv[i] ) ) {

			nodef= 1;

		} else {

			if ( '-' != argv[i][0] ) {
			
				filename= strdup( argv[i] );

			} else {

				fprintf( stderr, "ERROR: Unknown argument.\n" );
				exit(1);
			}
		}
	}


	if ( NULL == filename ) {
	
		printf( " no input file specified, abort\n" );
		exit(1);
	}
	
	if ( 1 > maxfilehandles ) {
	
		printf( " there must be at least 1 available filehandle\n" );
		exit(1);
	}
	
	/* open filemanager */
	manager= OTF_FileManager_open( maxfilehandles );
	assert( NULL != manager );

	/* Open OTF Reader */
	reader= OTF_Reader_open( filename, manager );
	if ( NULL == reader ) {

		fprintf( stderr, "%s ERROR: could not open '%s'\n", "otfaux", filename );
		return 1;
	}

	OTF_Reader_setBufferSizes( reader, buffersize );
	mc= OTF_Reader_getMasterControl( reader );


	/* check if there are already existing statistics and snapshots */
	/* get streamcount and namestub */
	namestub= OTF_stripFilename( filename );

	free( filename );

	Control control;
	control.num= 0;
	control.minNum= (uint64_t) minNum;
	control.maxNum= (uint64_t) maxNum;
	control.minTime= (uint64_t) minTime;
	control.maxTime= (uint64_t) maxTime;
	control.nodef= nodef;


	handlers= OTF_HandlerArray_open();

	OTF_HandlerArray_setHandler( handlers,
		(OTF_FunctionPointer*) handleDefFunction,
		OTF_DEFFUNCTION_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, (void*) &control,
		OTF_DEFFUNCTION_RECORD );

	OTF_HandlerArray_setHandler( handlers,
		(OTF_FunctionPointer*) handleDefTimerresolution,
		OTF_DEFTIMERRESOLUTION_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, (void*) &control,
		OTF_DEFTIMERRESOLUTION_RECORD );

	OTF_HandlerArray_setHandler( handlers,
		(OTF_FunctionPointer*) handleDefProcess,
		OTF_DEFPROCESS_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, (void*) &control,
		OTF_DEFPROCESS_RECORD );


	OTF_HandlerArray_setHandler( handlers,
		(OTF_FunctionPointer*) handleDefinitionComment, OTF_DEFINITIONCOMMENT_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, (void*) &control, OTF_DEFINITIONCOMMENT_RECORD );

	OTF_HandlerArray_setHandler( handlers,
		(OTF_FunctionPointer*) handleDefProcessGroup, OTF_DEFPROCESSGROUP_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, (void*) &control, OTF_DEFPROCESSGROUP_RECORD );

	OTF_HandlerArray_setHandler( handlers,
		(OTF_FunctionPointer*) handleDefFunctionGroup, OTF_DEFFUNCTIONGROUP_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, (void*) &control, OTF_DEFFUNCTIONGROUP_RECORD );

	OTF_HandlerArray_setHandler( handlers,
		(OTF_FunctionPointer*) handleDefCollectiveOperation, OTF_DEFCOLLOP_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, (void*) &control, OTF_DEFCOLLOP_RECORD );

	OTF_HandlerArray_setHandler( handlers,
		(OTF_FunctionPointer*) handleDefCounter, OTF_DEFCOUNTER_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, (void*) &control, OTF_DEFCOUNTER_RECORD );

	OTF_HandlerArray_setHandler( handlers,
		(OTF_FunctionPointer*) handleDefCounterGroup, OTF_DEFCOUNTERGROUP_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, (void*) &control, OTF_DEFCOUNTERGROUP_RECORD );

	OTF_HandlerArray_setHandler( handlers,
		(OTF_FunctionPointer*) handleDefScl, OTF_DEFSCL_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, (void*) &control, OTF_DEFSCL_RECORD );

	OTF_HandlerArray_setHandler( handlers,
		(OTF_FunctionPointer*) handleDefSclFile, OTF_DEFSCLFILE_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, (void*) &control, OTF_DEFSCLFILE_RECORD );

	OTF_HandlerArray_setHandler( handlers,
		(OTF_FunctionPointer*) handleDefCreator, OTF_DEFCREATOR_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, (void*) &control, OTF_DEFCREATOR_RECORD );


	
	OTF_HandlerArray_setHandler( handlers,
		(OTF_FunctionPointer*) handleCounter,
		OTF_COUNTER_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, (void*) &control,
		OTF_COUNTER_RECORD );

	OTF_HandlerArray_setHandler( handlers,
		(OTF_FunctionPointer*) handleEnter,
		OTF_ENTER_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, (void*) &control,
		OTF_ENTER_RECORD );

	OTF_HandlerArray_setHandler( handlers,
		(OTF_FunctionPointer*) handleRecvmsg,
		OTF_RECEIVE_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, (void*) &control,
		OTF_RECEIVE_RECORD );

	OTF_HandlerArray_setHandler( handlers,
		(OTF_FunctionPointer*) handleSendmsg,
		OTF_SEND_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, (void*) &control,
		OTF_SEND_RECORD );

	OTF_HandlerArray_setHandler( handlers,
		(OTF_FunctionPointer*) handleLeave,
		OTF_LEAVE_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, (void*) &control,
		OTF_LEAVE_RECORD );
	
	OTF_HandlerArray_setHandler( handlers, 
		(OTF_FunctionPointer*) handleBeginProcess,
		OTF_BEGINPROCESS_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, (void*) &control, 
		OTF_BEGINPROCESS_RECORD );

	OTF_HandlerArray_setHandler( handlers,
		(OTF_FunctionPointer*) handleEndProcess,
		OTF_ENDPROCESS_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, (void*) &control,
		OTF_ENDPROCESS_RECORD );

    OTF_HandlerArray_setHandler( handlers, 
    	(OTF_FunctionPointer*) handleCollectiveOperation,
    	OTF_COLLOP_RECORD );
    OTF_HandlerArray_setFirstHandlerArg( handlers, (void*) &control, 
    	OTF_COLLOP_RECORD );
		
	OTF_HandlerArray_setHandler( handlers, 
		(OTF_FunctionPointer*) handleEventComment,
		OTF_EVENTCOMMENT_RECORD );
	OTF_HandlerArray_setFirstHandlerArg( handlers, (void*) &control, 
		OTF_EVENTCOMMENT_RECORD );

	/* cout << "read " << read << " defs" << endl; */
	OTF_Reader_readDefinitions( reader, handlers );


	/* init read operation but do not start to read records yet. this ensures the
	time interval of the trace is extracted */
	OTF_Reader_setRecordLimit( reader, 0 );
	OTF_Reader_readEvents( reader, handlers );
	/* cout << "read " << read << " events" << endl; */
	OTF_Reader_setRecordLimit( reader, (uint64_t) -1 );

	uint64_t tmin;
	uint64_t tcur;
	uint64_t tmax;

	OTF_Reader_eventProgress( reader, &tmin, &tcur, &tmax );
	/*
	cout << "total time " << 
		(unsigned long long) tmin << " - " << 
		(unsigned long long) tmax << endl;
	*/

	/* restrict streams resp. processes to be read */

	OTF_Reader_readEvents( reader, handlers );

	OTF_Reader_close( reader );

	OTF_HandlerArray_close( handlers );
	OTF_FileManager_close( manager );

	return (0);
}


